--[[
Name: Banzai-1.0
Revision: $Rev: 19044 $
Author(s): Rabbit (rabbit.magtheridon@gmail.com), maia
Documentation: http://www.wowace.com/index.php/Banzai-1.0_API_Documentation
SVN: http://svn.wowace.com/root/trunk/BanzaiLib/Banzai-1.0
Description: Aggro notification library.
Dependencies: AceLibrary, AceEvent-2.0, RosterLib-2.0
]]

-------------------------------------------------------------------------------
-- Locals
-------------------------------------------------------------------------------

local MAJOR_VERSION = "Banzai-1.0"
local MINOR_VERSION = "$Revision: 19044 $"

if not AceLibrary then error(MAJOR_VERSION .. " requires AceLibrary") end
if not AceLibrary:IsNewVersion(MAJOR_VERSION, MINOR_VERSION) then return end
if not AceLibrary:HasInstance("RosterLib-2.0") then error(MAJOR_VERSION .. " requires RosterLib-2.0.") end

local lib = {}
AceLibrary("AceEvent-2.0"):embed(lib)

local RL = nil
local roster = nil
local playerName = nil
local running = nil

-------------------------------------------------------------------------------
-- Initialization
-------------------------------------------------------------------------------

-- Activate a new instance of this library
function activate(self, oldLib, oldDeactivate)
	if oldLib then
		self.vars = oldLib.vars
		if oldLib:IsEventScheduled("UpdateAggroList") then
			oldLib:CancelScheduledEvent("UpdateAggroList")
		end
	end

	RL = AceLibrary("RosterLib-2.0")
	roster = RL.roster
	playerName = UnitName("player")

	self:RegisterEvent("AceEvent_EventRegistered", "StartOrStop")
	self:RegisterEvent("AceEvent_EventUnregistered", "StartOrStop")

	if oldDeactivate then oldDeactivate(oldLib) end
end

-------------------------------------------------------------------------------
-- Events
-------------------------------------------------------------------------------

function lib:StartOrStop()
	local aceEvent = AceLibrary("AceEvent-2.0")
	if aceEvent:IsEventRegistered("Banzai_UnitGainedAggro") or
		aceEvent:IsEventRegistered("Banzai_PlayerGainedAggro") or
		aceEvent:IsEventRegistered("Banzai_UnitLostAggro") or
		aceEvent:IsEventRegistered("Banzai_PlayerLostAggro") or
		aceEvent:IsEventRegistered("Banzai_Run") then
		if not self:IsEventScheduled("UpdateAggroList") then
			self:ScheduleRepeatingEvent("UpdateAggroList", self.UpdateAggroList, 0.2, self)
			running = true
			self:TriggerEvent("Banzai_Enabled")
		end
	elseif self:IsEventRegistered("UpdateAggroList") then
		self:CancelScheduledEvent("UpdateAggroList")
		for i, unit in pairs(roster) do
			unit.banzai = nil
			unit.banzaiModifier = nil
		end
		running = nil
		self:TriggerEvent("Banzai_Disabled")
	end
end

-------------------------------------------------------------------------------
-- Library
-------------------------------------------------------------------------------

function lib:UpdateAggroList()
	local oldBanzai = {}

	for i, unit in pairs(roster) do
		oldBanzai[unit.unitid] = unit.banzai

		-- deduct aggro for all, increase it later for everyone with aggro
		if not unit.banzaiModifier then unit.banzaiModifier = 0 end
		unit.banzaiModifier = math.max(0, unit.banzaiModifier - 5)

		-- check for aggro
		local targetId = unit.unitid .. "target"
		local targetName = UnitName(targetId .. "target")
		if roster[targetName] and UnitCanAttack("player", targetId) and UnitCanAttack(targetId, "player") then
			if not roster[targetName].banzaiModifier then roster[targetName].banzaiModifier = 0 end
			roster[targetName].banzaiModifier = roster[targetName].banzaiModifier + 10
			if not roster[targetName].banzaiTarget then roster[targetName].banzaiTarget = {} end
			table.insert(roster[targetName].banzaiTarget, targetId)
		end

		-- cleanup
		unit.banzaiModifier = math.max(0, unit.banzaiModifier)
		unit.banzaiModifier = math.min(25, unit.banzaiModifier)

		-- set aggro
		unit.banzai = (unit.banzaiModifier > 15)
	end

	for i, unit in pairs(roster) do
		if oldBanzai[unit.unitid] ~= nil and oldBanzai[unit.unitid] ~= unit.banzai then
			-- Aggro status has changed.
			if unit.banzai == true and unit.banzaiTarget then
				-- Unit has aggro
				self:TriggerEvent("Banzai_UnitGainedAggro", unit.unitid, unit.banzaiTarget)
				if unit.name == playerName then
					self:TriggerEvent("Banzai_PlayerGainedAggro", unit.banzaiTarget)
				end
			elseif unit.banzai == false then
				-- Unit lost aggro
				self:TriggerEvent("Banzai_UnitLostAggro", unit.unitid)
				if unit.name == playerName then
					self:TriggerEvent("Banzai_PlayerLostAggro", unit.unitid)
				end
			end
		end
		unit.banzaiTarget = nil
	end

	oldBanzai = nil
end

-------------------------------------------------------------------------------
-- API
-------------------------------------------------------------------------------

function lib:GetUnitAggroByUnitId( unitId )
	if not running then error("Banzai-1.0 is not running. You must register for one of the events.") end
	local rosterUnit = RL:GetUnitObjectFromUnit(unitId)
	if not rosterUnit then return nil end
	return rosterUnit.banzai
end

function lib:GetUnitAggroByUnitName( unitName )
	if not running then error("Banzai-1.0 is not running. You must register for one of the events.") end
	local rosterUnit = RL:GetUnitObjectFromName(unitName)
	if not rosterUnit then return nil end
	return rosterUnit.banzai
end

-------------------------------------------------------------------------------
-- Register
-------------------------------------------------------------------------------
AceLibrary:Register(lib, MAJOR_VERSION, MINOR_VERSION, activate)

