﻿local L = AceLibrary("AceLocale-2.2"):new("sRaidFrames")
local BS = AceLibrary("Babble-Spell-2.2")
local proximity = ProximityLib:GetInstance("1")
local surface = AceLibrary("Surface-1.0") 
local roster = AceLibrary("RosterLib-2.0")

surface:Register("Otravi", "Interface\\AddOns\\sRaidFrames\\textures\\otravi")
surface:Register("Smooth", "Interface\\AddOns\\sRaidFrames\\textures\\smooth")
surface:Register("Striped", "Interface\\AddOns\\sRaidFrames\\textures\\striped")
surface:Register("BantoBar", "Interface\\AddOns\\sRaidFrames\\textures\\bantobar")

local math_mod = math.fmod or math.mod 

sRaidFrames = AceLibrary("AceAddon-2.0"):new(
	"AceDB-2.0",
	"AceEvent-2.0",
	"AceConsole-2.0",
	"AceHook-2.1",
	"FuBarPlugin-2.0"
)

-- Look I'm a fubar plugin!!
sRaidFrames.hasIcon = "Interface\\Icons\\INV_Helmet_06"
sRaidFrames.defaultMinimapPosition = 180
sRaidFrames.cannotDetachTooltip = true
sRaidFrames.hasNoColor = true
sRaidFrames.clickableTooltip = false
sRaidFrames.hideWithoutStandby = true
sRaidFrames.independentProfile = true

-- bindings
BINDING_NAME_ShowHideRaidWindows = L["Show/Hide Raid Windows"]
BINDING_NAME_ToggleSort = L["Toggle Sort"]
BINDING_NAME_Sortbyclass = L["Sort by class"]
BINDING_NAME_Sortbygroup = L["Sort by group"]
BINDING_NAME_ToggleBuffDebuffview = L["Toggle Buff/Debuff view"]
BINDING_NAME_Toggledisplayofonlydispellabledebuffs = L["Toggle display of only dispellable debuffs"]

function sRaidFrames:OnInitialize()

	self:RegisterDB("sRaidFramesDB")
	self:Variables()

	self:RegisterDefaults("profile", {
		Locked				= false,
		SortBy				= "class",
		HealthFormat	= 'percent',
		Invert 				= false,
		Scale					= 1,
		Border				= true,
		Texture				= "Otravi",
		BuffType			= "debuffs",
		ShowOnlyDispellable	= 1,
		BackgroundColor		= {r = 0.47, g = 0.72, b = 1, a = 0.7},
		BorderColor		= {r = 1, g = 1, b = 1, a = 1},
		Growth				= "down",
		Spacing				= 0,
		ShowGroupTitles		= true,
		SubSort				= "name",
		UnitTooltipMethod		= "notincombat",
		BuffTooltipMethod = "always",
		DebuffTooltipMethod = "always",
		ClassFilter			= {["WARRIOR"] = true, ["PALADIN"] = true, ["SHAMAN"] = true, ["HUNTER"] = true, ["WARLOCK"] = true, ["MAGE"] = true, ["DRUID"] = true, ["ROGUE"] = true, ["PRIEST"] = true},
		GroupFilter			= {true, true, true, true, true, true, true, true},
		BuffFilter			= {},
		PowerFilter			= {[0] = true,[1] = true,[2] = true,[3] = true},
		AggroCheck				= false,
		RangeCheck 			= false,
		RangeFrequency 		= 2,
		RangeAlpha 			= 0.5,
		HighlightTarget	= false,
	})

	self:RegisterChatCommand({"/srf", "/sraidframes"}, self.options)

	self.opt = self.db.profile

	self.OnMenuRequest = self.options

	self.master = CreateFrame("Frame", "sRaidFrame", UIParent)
	self.master:SetMovable(true)
	self.master:SetScale(self.opt.Scale)

	self.master:SetHeight(200);
	self.master:SetWidth(200);
end

function sRaidFrames:OnProfileEnable()
	self.opt = self.db.profile
	self:SetPosition()
end

function sRaidFrames:OnEnable()	
	self:chatUpdateBuffMenu()

	self:SecureHook("SecureRaidGroupHeader_Update", "SecureHeaderUpdate")
	self:RegisterBucketEvent("RAID_ROSTER_UPDATE", 0.1, "UpdateRoster")

	for i = 1, 8 do
		self:CreateGroupFrame(i)
	end
	
	self:SetGroupFilters()
	self:SetSortMethods()
	self:SetGrowth()
	self:SetPosition()

	self:UpdateRoster()
end

function sRaidFrames:OnDisable()
	self.master:Hide()
end

function sRaidFrames:JoinedRaid()
	self:Print("Joined a raid, enabling raid frames")
	self.enabled = true

	self:RegisterBucketEvent("UNIT_HEALTH", 0.2, "UpdateUnit")
	self:RegisterBucketEvent("UNIT_AURA", 0.2, "UpdateBuffs")
	
	self:RegisterEvent("PLAYER_TARGET_CHANGED", "UpdateTarget")

	self:RegisterEvent("Banzai_UnitGainedAggro")
	self:RegisterEvent("Banzai_UnitLostAggro")
	
	self:RegisterEvent("oRA_PlayerCanResurrect")
	self:RegisterEvent("oRA_PlayerResurrected")
	self:RegisterEvent("oRA_PlayerNotResurrected")

	-- TODO: only updateunit
	self:ScheduleRepeatingEvent("sRaidFramesUpdateAll", self.UpdateAll, 1.5, self)
	self:ScheduleRepeatingEvent("sRaidFramesRangeCheck", self.RangeCheck, self.opt.RangeFrequency, self)

	self:UpdateRoster()
	self:UpdateAll()

	self.master:Show()
end

function sRaidFrames:LeftRaid()
	self:Print("Left raid, disabling raid frames")
	self.visible = {}
	self.enabled = false

	self:UnregisterBucketEvent("UNIT_HEALTH")
	self:UnregisterBucketEvent("UNIT_AURA")

	self:UnregisterEvent("Banzai_UnitGainedAggro")
	self:UnregisterEvent("Banzai_UnitLostAggro")
	
	self:UnregisterEvent("oRA_PlayerCanResurrect")
	self:UnregisterEvent("oRA_PlayerResurrected")
	self:UnregisterEvent("oRA_PlayerNotResurrected")

	self:CancelScheduledEvent("sRaidFramesUpdateAll")
	self:CancelScheduledEvent("sRaidFramesRangeCheck")

	self.master:Hide()
end

function sRaidFrames:UpdateAll()
	self:UpdateUnit(self.visible)
	self:UpdateBuffs(self.visible)
end

function sRaidFrames:Variables()
	self.enabled = false
	self.frames, self.visible, self.groupframes = {}, {}, {}
	self.unavail, self.res = {}, {}, {}
	self.inCombat = false

	self.debuffColors = {}
	self.debuffColors["Curse"]    = { ["r"] = 1, ["g"] = 0, ["b"] = 0.75, ["a"] = 0.5, ["priority"] = 4 }
	self.debuffColors["Magic"]    = { ["r"] = 1, ["g"] = 0, ["b"] = 0, ["a"] = 0.5, ["priority"] = 3 }
	self.debuffColors["Disease"]  = { ["r"] = 1, ["g"] = 1, ["b"] = 0, ["a"] = 0.5, ["priority"] = 2 }
	self.debuffColors["Poison"]   = { ["r"] = 0, ["g"] = 0.5, ["b"] = 0, ["a"] = 0.5, ["priority"] = 1 }
	
	self.cooldownSpells = {}
	self.cooldownSpells["WARLOCK"] = BS["Soulstone Resurrection"]
	self.cooldownSpells["DRUID"] = BS["Rebirth"]
	self.cooldownSpells["SHAMAN"] = BS["Reincarnation"]
	
	self.classGroups = {}
	self.classGroups[1] = "WARRIOR"
	self.classGroups[2] = "MAGE"
	self.classGroups[3] = (UnitFactionGroup("player") == "Alliance") and "PALADIN" or "SHAMAN"
	self.classGroups[4] = "DRUID"
	self.classGroups[5] = "HUNTER"
	self.classGroups[6] = "ROGUE"
	self.classGroups[7] = "WARLOCK"
	self.classGroups[8] = "PRIEST"
	
	self.eclass2class = {}
	self.eclass2class["PALADIN"] = L["Paladin"]
	self.eclass2class["WARRIOR"] = L["Warrior"]
	self.eclass2class["SHAMAN"] = L["Shaman"]
	self.eclass2class["MAGE"] = L["Mage"]
	self.eclass2class["HUNTER"] = L["Hunter"]
	self.eclass2class["DRUID"] = L["Druid"]
	self.eclass2class["ROGUE"] = L["Rogue"]
	self.eclass2class["PRIEST"] = L["Priest"]
	self.eclass2class["WARLOCK"] = L["Warlock"]
	

	self.ManaBarColor = ManaBarColor

	-- Nurfed mana color, where is the love :(
	--self.ManaBarColor[0] = { r = 0.00, g = 1.00, b = 1.00, prefix = TEXT(MANA) };
	
	self.statusmaps = {}
	self.statusmaps["Aggro"] = {["priority"] = 90}
	self.statusmaps["Target"] = {["priority"] = 50}
	
	self.statusstate = {}
end

function sRaidFrames:UpdateRoster()
	local num = GetNumRaidMembers()

	if num == 0 then
		if self.enabled then
			self:LeftRaid()
		end
		return
	end

	if not self.enabled then
			self:JoinedRaid()
	end

	self:UpdateVisibility()
end

function sRaidFrames:QueryVisibility(id)
	local unitid = "raid" .. id
	if not UnitName(unitid) then
		return false
	end

	local _, eClass = UnitClass(unitid)
	if not self.opt.ClassFilter[eClass] then
		return false
	end

	local _, _, subgroup = GetRaidRosterInfo(id)
	if not self.opt.GroupFilter[subgroup] then
		return false
	end

	return true
end

function sRaidFrames:UpdateVisibility()
	for id = 1, MAX_RAID_MEMBERS do
		if self:QueryVisibility(id) then
			if not self.visible["raid" .. id] then
				self.visible["raid" .. id] = true;
			end
		else
			if self.visible["raid" .. id] then
				self.visible["raid" .. id] = nil;
			end
		end
	end
	
	self:UpdateAll()
end

function sRaidFrames:UpdateTarget()
	if not self.opt.HighlightTarget then return end
	
	for unit in pairs(self.visible) do
		if UnitIsUnit(unit, "target") then
			self:SetStatusBorder("Target", unit, {r = 1, g = 0.75, b = 0})
		else
			self:UnsetStatusBorder("Target", unit)
		end
	end
end

function sRaidFrames:Banzai_UnitGainedAggro(unit, unitTable)
	if not unit or not self.visible[unit] or not self.opt.AggroCheck then return end

	self:SetStatusBorder("Aggro", unit, {r = 1, g = 0, b = 0})
end

function sRaidFrames:Banzai_UnitLostAggro(unit)
	if not unit or not self.visible[unit] or not self.opt.AggroCheck then return end

	self:UnsetStatusBorder("Aggro", unit)
end

function sRaidFrames:oRA_PlayerCanResurrect(msg, player)
	local unit = roster:GetUnitIDFromName(player)
	if unit then self.res[unit] = 1 end
end

function sRaidFrames:oRA_PlayerResurrected(msg, player)
	local unit = roster:GetUnitIDFromName(player)
	if unit then self.res[unit] = 2 end
end

function sRaidFrames:oRA_PlayerNotResurrected(msg, player)
	local unit = roster:GetUnitIDFromName(player)
	if unit then self.res[unit] = nil end
end

function sRaidFrames:RangeCheck()
	if not self.opt.RangeCheck then return end

    local now = GetTime()

    for unit in pairs(self.visible) do
		local _, time = proximity:GetUnitRange(unit)
		local seen = now - (time or 100)

		if time and seen < 3 then
			for _, f in pairs(self:FindUnitFrames(unit)) do
				f:SetAlpha(1)
			end
		else
			for _, f in pairs(self:FindUnitFrames(unit)) do
				f:SetAlpha(self.opt.RangeAlpha)
			end
		end
	end
end

function sRaidFrames:UpdateRangeFrequency()
		self:CancelScheduledEvent("sRaidFramesRangeCheck")
		self:ScheduleRepeatingEvent("sRaidFramesRangeCheck", self.RangeCheck, self.opt.RangeFrequency, self)
end

function sRaidFrames:UpdateUnit(units)
	for unit in pairs(units) do
		if self.visible[unit] and UnitExists(unit) then
			for _, f in pairs(self:FindUnitFrames(unit)) do
				local _, class = UnitClass(unit)
				if class then
					f.title:SetText(UnitName(unit))
					
					local color = RAID_CLASS_COLORS[class]
					f.title:SetTextColor(color.r, color.g, color.b, 1)
				else
					f.title:SetText(UnitName(unit) or L["Unknown"])
				end
	
				local feign = nil
	
				-- Silly hunters, why do you have to be so annoying
				if class == "HUNTER" then
					for i=1,32 do
						local texture = UnitBuff(unit, i)
						if not texture then break end
						if texture == "Interface\\Icons\\Ability_Rogue_FeignDeath" then
							feign = true
							break
						end
					end
				end
	
				if feign then
					f.hpbar.text:SetText("|cff00ff00"..L["Feign Death"].."|r")
					f.hpbar:SetValue(100)
					f.hpbar:SetStatusBarColor(0, 0.9, 0.5)
					f.mpbar:SetValue(0)
				else
					local status, dead, ghost = nil, UnitIsDead(unit), UnitIsGhost(unit)
					
					if not UnitIsConnected(unit) then status = "|cffff0000"..L["Offline"].."|r"
					elseif self.res[unit] == 1 and dead then status = "|cffff8c00"..L["Can Recover"].."|r"
					elseif self.res[unit] == 2 and (dead or ghost) then status = "|cffff8c00"..L["Resurrected"].."|r"
					elseif ghost then status = "|cffff0000"..L["Released"].."|r"
					elseif dead then status = "|cffff0000"..L["Dead"].."|r"
					end
	
					if status then
						self.unavail[unit] = true
						f.hpbar.text:SetText(status)
						f.hpbar:SetValue(0)
						f.mpbar.text:SetText()
						f.mpbar:SetValue(0)
						f:SetBackdropColor(0.3, 0.3, 0.3, 1)
					else
						self.unavail[unit] = false
						self.res[unit] = nil
						local hp = UnitHealth(unit) or 0
						local hpmax = UnitHealthMax(unit)
						local hpp = (hpmax ~= 0) and ceil((hp / hpmax) * 100) or 0
						local hptext, hpvalue = nil, nil
	
						if self.opt.HealthFormat == "percent" then
							hptext = hpp .."%"
						elseif self.opt.HealthFormat == "deficit" then
							hptext = (hp-hpmax) ~=  0 and (hp-hpmax) or nil
						elseif self.opt.HealthFormat == "current" then
							hptext = hp
						elseif self.opt.HealthFormat == "curmax" then
							hptext = hp .."/".. hpmax
						end
						
						if self.opt.Invert then
							hpvalue = 100 - hpp
						else
							hpvalue = hpp
						end
	
						f.hpbar.text:SetText(hptext)
						f.hpbar:SetValue(hpvalue)
						f.hpbar:SetStatusBarColor(self:GetHPSeverity(hpp/100))
	
						local mp = UnitMana(unit) or 0
						local mpmax = UnitManaMax(unit)
						local mpp = (mpmax ~= 0) and ceil((mp / mpmax) * 100) or 0
	
						local powerType = UnitPowerType(unit)
						if self.opt.PowerFilter[powerType] == false then
							f.mpbar:SetValue(0)
						else
							local color = self.ManaBarColor[powerType]
							f.mpbar:SetStatusBarColor(color.r, color.g, color.b)
							f.mpbar:SetValue(mpp)
						end
					end
				end
			end
		end
	end
end

function sRaidFrames:UpdateBuffs(units)
	for unit in pairs(units) do
		if self.visible[unit] and UnitExists(unit) then
			for _, f in pairs(self:FindUnitFrames(unit)) do
				local cAura = nil
	
				for i = 1, 2 do
					f["aura".. i]:Hide()
				end
	
				for i = 1, 4 do
					f["buff".. i]:Hide()
				end
	
				local debuffSlots = 0
				for i=1,16 do
					local debuffName, _, debuffTexture, debuffApplications, debuffType = UnitDebuff(unit, i, self.opt.ShowOnlyDispellable)
					if not debuffTexture then break end
	
					if debuffType ~= nil and self.debuffColors[debuffType] and ((cAura and cAura.priority < self.debuffColors[debuffType].priority) or not cAura) then
						cAura = self.debuffColors[debuffType]
					end
	
					if (self.opt.BuffType == "debuffs" or self.opt.BuffType == "buffsifnotdebuffed") and debuffSlots < 2 then
						debuffSlots = debuffSlots + 1
						local debuffFrame = f["aura".. debuffSlots]
						debuffFrame.unitid = unit
						debuffFrame.debuffid = i
						debuffFrame.showDispellable = self.opt.ShowOnlyDispellable
						debuffFrame.count:SetText(debuffApplications > 1 and debuffApplications or nil);
						debuffFrame.texture:SetTexture(debuffTexture)
						debuffFrame:Show()
					end
				end
	
				if cAura then
					f:SetBackdropColor(cAura.r, cAura.g, cAura.b, cAura.a);
				elseif not self.unavail[unit] then
					f:SetBackdropColor(self.opt.BackgroundColor.r, self.opt.BackgroundColor.g, self.opt.BackgroundColor.b, self.opt.BackgroundColor.a)
				end
	
				f.mpbar.text:SetText()
	
				for i=1,32 do
					local buffName, _, buffTexture = UnitBuff(unit, i)
					if not buffTexture then break end
	
					if buffTexture == "Interface\\Icons\\Spell_Nature_TimeStop" and buffName == BS["Divine Intervention"] then
						f.hpbar.text:SetText("|cffff0000"..L["Intervened"].."|r")
					elseif buffTexture == "Interface\\Icons\\Spell_Nature_Lightning" and buffName == BS["Innervate"] then
						f.mpbar.text:SetText("|cff00ff00"..L["Innervating"].."|r")
					elseif buffTexture == "Interface\\Icons\\Spell_Holy_GreaterHeal" and buffName == BS["Spirit of Redemption"] then
						f.hpbar.text:SetText("|cffff0000"..L["Spirit"].."|r")
					elseif buffTexture == "Interface\\Icons\\Ability_Warrior_ShieldWall" and buffName == BS["Shield Wall"] then
						f.mpbar.text:SetText("|cffffffff"..BS["Shield Wall"].."|r")
					elseif buffTexture == "Interface\\Icons\\Spell_Holy_AshesToAshes" and buffName == BS["Last Stand"] then
						f.mpbar.text:SetText("|cffffffff"..BS["Last Stand"].."|r")
					elseif buffTexture == "Interface\\Icons\\INV_Misc_Gem_Pearl_05" then
						f.mpbar.text:SetText("|cffffffff"..L["Gift of Life"].."|r")
					elseif buffTexture == "Interface\\Icons\\Spell_Frost_Frost" and buffName == BS["Ice Block"] then
						f.mpbar.text:SetText("|cffbfefff"..BS["Ice Block"].."|r")
					elseif buffTexture == "Interface\\Icons\\Spell_Holy_SealOfProtection" and buffName == BS["Blessing of Protection"] then
						f.mpbar.text:SetText("|cffffffff"..L["Protection"].."|r")
					elseif buffTexture == "Interface\\Icons\\Spell_Holy_DivineIntervention" and buffName == BS["Divine Shield"] then
						f.mpbar.text:SetText("|cffffffff"..BS["Divine Shield"].."|r")
					elseif buffTexture == "Interface\\Icons\\Ability_Vanish" and buffName == BS["Vanish"] then
						f.mpbar.text:SetText("|cffffffff"..L["Vanished"].."|r")
					elseif buffTexture == "Interface\\Icons\\Ability_Stealth" and buffName == BS["Stealth"] then
						f.mpbar.text:SetText("|cffffffff"..L["Stealthed"].."|r")
					elseif buffTexture == "Interface\\Icons\\Spell_Holy_PowerInfusion" and buffName == BS["Power Infusion"] then
						f.mpbar.text:SetText("|cffffffff"..L["Infused"].."|r")
					elseif buffTexture == "Interface\\Icons\\Spell_Holy_Excorcism" and buffName == BS["Fear Ward"] then
						f.mpbar.text:SetText("|cffffff00"..BS["Fear Ward"].."|r")
		--			elseif buffTexture == "Interface\\Icons\\Spell_Holy_Renew" and buffName == BS["Renew"] then
		--				f.mpbar.text:SetText("|cff00ff00"..BS["Renew"].."|r")
					elseif buffTexture == "Interface\\Icons\\Spell_Shadow_Shadowform" and buffName == BS["Shadowform"] then
						f.mpbar.text:SetText("|cff800080"..BS["Shadowform"].."|r")
					end
				end
	
	
				if self.opt.BuffType == "buffs" or (self.opt.BuffType == "buffsifnotdebuffed" and debuffSlots == 0) then
					local buffSlots = 0
					local showOnlyCastable = 1
					if next(self.opt.BuffFilter) then
						showOnlyCastable = 0
					end
					for i=1,32 do
						local buffName, _, buffTexture, buffApplications = UnitBuff(unit, i, showOnlyCastable)
						if not buffTexture then break end
	
						if showOnlyCastable == 1 or self.opt.BuffFilter[buffName] then
							buffSlots = buffSlots + 1
							local buffFrame = f["buff".. buffSlots]
							buffFrame.buffid = i
							buffFrame.unitid = unit
							buffFrame.showCastable = showOnlyCastable
							buffFrame.count:SetText(buffApplications > 1 and buffApplications or nil)
							buffFrame.texture:SetTexture(buffTexture)
							buffFrame:Show()
						end
	
						if buffSlots == 4 then break end
					end
				end
			end
		end
	end
end

function sRaidFrames:SetStatusText(type, unit, text, color)
	for _, f in pairs(self:FindUnitFrames(unit)) do
		f:SetText(text, color.r, color.g, color.b, color.a)
	end
end

function sRaidFrames:SetStatusBorder(type, unit, color)
	if self.statusmaps[type] and self.statusstate[unit] and self.statusmaps[type].priority < self.statusstate[unit].priority then return end
	if not color.a then color.a = self.opt.BorderColor.a end
	
	for _, f in pairs(self:FindUnitFrames(unit)) do
		f:SetBackdropBorderColor(color.r, color.g, color.b, color.a)
	end
	self.statusstate[unit] = self.statusmaps[type]
end

function sRaidFrames:UnsetStatusBorder(type, unit)
	if self.statusmaps[type] and self.statusstate[unit] and self.statusmaps[type].priority < self.statusstate[unit].priority then return end
	
	for _, f in pairs(self:FindUnitFrames(unit)) do
		f:SetBackdropBorderColor(self.opt.BorderColor.r, self.opt.BorderColor.g, self.opt.BorderColor.b, self.opt.BorderColor.a)
	end
	self.statusstate[unit] = nil
end

function sRaidFrames:GetHPSeverity(percent)
	if (percent >= 0.5) then
		return (1.0-percent)*2, 1.0, 0.0
	else
		return 1.0, percent*2, 0.0
	end
end

function sRaidFrames:QueryTooltipDisplay(value)
	if value == "never" then
		return false
	elseif value == "notincombat" and UnitAffectingCombat("player") then
		return false
	else
		return true
	end
end

function sRaidFrames:UnitTooltip(frame)
	local name, rank, subgroup, level, class, fileName, zone, online, isDead = GetRaidRosterInfo(frame.id)
	GameTooltip:SetOwner(frame)
	GameTooltip:AddDoubleLine(name, level, RAID_CLASS_COLORS[fileName].r, RAID_CLASS_COLORS[fileName].g, RAID_CLASS_COLORS[fileName].b, 1, 1, 1)
	if UnitIsAFK(frame.unit) then
		GameTooltip:AddLine("AFK: Away From Keyboard", 1, 1, 0)
	end
	if UnitIsDND(frame.unit) then
		GameTooltip:AddLine("DND: Do Not Disturb", 1, 1, 0)
	end
	GameTooltip:AddLine(UnitRace(frame.unit) .. " " .. class, 1, 1, 1);
	GameTooltip:AddDoubleLine(zone, "Group ".. subgroup, 1, 1, 1, 1, 1, 1);


	if oRAOCoolDown and oRAOCoolDown.db.profile.cooldowns and oRAOCoolDown.db.profile.cooldowns[name] and self.cooldownSpells[fileName] then
		GameTooltip:AddLine(" ")
		local expire = oRAOCoolDown.db.profile.cooldowns[name]-time()
		if expire > 0 then
  		GameTooltip:AddDoubleLine(self.cooldownSpells[fileName], SecondsToTime(expire), nil, nil, nil, 1, 0, 0)
  	else
  		GameTooltip:AddDoubleLine(self.cooldownSpells[fileName], "Ready!", nil, nil, nil, 0, 1, 0)
  	end
	end

	GameTooltip:Show()
end

function sRaidFrames_initialHeaderConfigFunction(frame)
	frame:SetAttribute("initial-width", 90)
	frame:SetAttribute("initial-height", 40)
	frame:SetAttribute("type1", "target")
end

function sRaidFrames_InitUnitFrame(frame)
	sRaidFrames:CreateUnitFrame(frame)
end

function sRaidFrames_OnAttributeChanged(self, name, value)
	if name == "unit" then
		if value then
			_, _, self.id = string.find(value, "(%d+)") -- TODO, don't know if I like this
			self.unit = value
		else
			self.id = nil
			self.unit = nil
		end
	end
end

-- Adapts frames created by Secure Headers
function sRaidFrames:CreateUnitFrame(f)
	f:EnableMouse(true)
	f:RegisterForClicks("LeftButtonUp", "RightButtonUp", "MiddleButtonUp", "Button4Up", "Button5Up")
	
	f:SetScript("OnEnter", function() if self:QueryTooltipDisplay(self.opt.UnitTooltipMethod) then self:UnitTooltip(this) end end)
	f:SetScript("OnLeave", function() GameTooltip:Hide() end)

	f.title = f:CreateFontString(nil, "ARTWORK")
	f.title:SetFontObject(GameFontNormalSmall)
	f.title:SetJustifyH("LEFT")

	f.aura1 = CreateFrame("Button", nil, f)
	f.aura1:SetScript("OnEnter", function() if self:QueryTooltipDisplay(self.opt.DebuffTooltipMethod) then GameTooltip:SetOwner(this) GameTooltip:SetUnitDebuff(this.unitid, this.debuffid, this.showDispellable) end end);
	f.aura1:SetScript("OnLeave", function() GameTooltip:Hide() end)
	f.aura1.texture = f.aura1:CreateTexture(nil, "ARTWORK")
	f.aura1.texture:SetAllPoints(f.aura1);
	f.aura1.count = f.aura1:CreateFontString(nil, "OVERLAY")
	f.aura1.count:SetFontObject(GameFontHighlightSmallOutline)
	f.aura1.count:SetJustifyH("CENTER")
	f.aura1.count:SetPoint("CENTER", f.aura1, "CENTER", 0, 0);
	f.aura1:Hide()

	f.aura2 = CreateFrame("Button", nil, f)
	f.aura2:SetScript("OnEnter", function() if self:QueryTooltipDisplay(self.opt.DebuffTooltipMethod) then GameTooltip:SetOwner(this) GameTooltip:SetUnitDebuff(this.unitid, this.debuffid, this.showDispellable) end end);
	f.aura2:SetScript("OnLeave", function() GameTooltip:Hide() end)
	f.aura2.texture = f.aura2:CreateTexture(nil, "ARTWORK")
	f.aura2.texture:SetAllPoints(f.aura2)
	f.aura2.count = f.aura2:CreateFontString(nil, "OVERLAY")
	f.aura2.count:SetFontObject(GameFontHighlightSmallOutline)
	f.aura2.count:SetJustifyH("CENTER")
	f.aura2.count:SetPoint("CENTER", f.aura2, "CENTER", 0, 0);
	f.aura2:Hide()

	f.buff1 = CreateFrame("Button", nil, f)
	f.buff1:SetScript("OnEnter", function() if self:QueryTooltipDisplay(self.opt.BuffTooltipMethod) then GameTooltip:SetOwner(this) GameTooltip:SetUnitBuff(this.unitid, this.buffid, this.showCastable) end end)
	f.buff1:SetScript("OnLeave", function() GameTooltip:Hide() end)
	f.buff1.texture = f.buff1:CreateTexture(nil, "ARTWORK")
	f.buff1.texture:SetAllPoints(f.buff1)
	f.buff1.count = f.buff1:CreateFontString(nil, "OVERLAY")
	f.buff1.count:SetFontObject(GameFontHighlightSmallOutline)
	f.buff1.count:SetJustifyH("CENTER")
	f.buff1.count:SetPoint("CENTER", f.buff1, "CENTER", 0, 0);
	f.buff1:Hide()

	f.buff2 = CreateFrame("Button", nil, f)
	f.buff2:SetScript("OnEnter", function() if self:QueryTooltipDisplay(self.opt.BuffTooltipMethod) then GameTooltip:SetOwner(this) GameTooltip:SetUnitBuff(this.unitid, this.buffid, this.showCastable) end end)
	f.buff2:SetScript("OnLeave", function() GameTooltip:Hide() end)
	f.buff2.texture = f.buff2:CreateTexture(nil, "ARTWORK")
	f.buff2.texture:SetAllPoints(f.buff2)
	f.buff2.count = f.buff2:CreateFontString(nil, "OVERLAY")
	f.buff2.count:SetFontObject(GameFontHighlightSmallOutline)
	f.buff2.count:SetJustifyH("CENTER")
	f.buff2.count:SetPoint("CENTER", f.buff2, "CENTER", 0, 0);
	f.buff2:Hide()

	f.buff3 = CreateFrame("Button", nil, f)
	f.buff3:SetScript("OnEnter", function() if self:QueryTooltipDisplay(self.opt.BuffTooltipMethod) then GameTooltip:SetOwner(this) GameTooltip:SetUnitBuff(this.unitid, this.buffid, this.showCastable) end end)
	f.buff3:SetScript("OnLeave", function() GameTooltip:Hide() end)
	f.buff3.texture = f.buff3:CreateTexture(nil, "ARTWORK")
	f.buff3.texture:SetAllPoints(f.buff3)
	f.buff3.count = f.buff3:CreateFontString(nil, "OVERLAY")
	f.buff3.count:SetFontObject(GameFontHighlightSmallOutline)
	f.buff3.count:SetJustifyH("CENTER")
	f.buff3.count:SetPoint("CENTER", f.buff3, "CENTER", 0, 0);
	f.buff3:Hide()

	f.buff4 = CreateFrame("Button", nil, f)
	f.buff4:SetScript("OnEnter", function() if self:QueryTooltipDisplay(self.opt.BuffTooltipMethod) then GameTooltip:SetOwner(this) GameTooltip:SetUnitBuff(this.unitid, this.buffid, this.showCastable) end end)
	f.buff4:SetScript("OnLeave", function() GameTooltip:Hide() end)
	f.buff4.texture = f.buff4:CreateTexture(nil, "ARTWORK")
	f.buff4.texture:SetAllPoints(f.buff4)
	f.buff4.count = f.buff4:CreateFontString(nil, "OVERLAY")
	f.buff4.count:SetFontObject(GameFontHighlightSmallOutline)
	f.buff4.count:SetJustifyH("CENTER")
	f.buff4.count:SetPoint("CENTER", f.buff4, "CENTER", 0, 0);
	f.buff4:Hide()

	f.hpbar = CreateFrame("StatusBar", nil, f)
	f.hpbar:SetStatusBarTexture(surface:Fetch(self.opt.Texture))
	f.hpbar:SetMinMaxValues(0,100)

	f.hpbar.texture = f.hpbar:CreateTexture(nil, "BORDER")
	f.hpbar.texture:SetTexture(surface:Fetch(self.opt.Texture))
	f.hpbar.texture:SetVertexColor(1, 0, 0, 0)

	f.hpbar.text = f.hpbar:CreateFontString(nil, "ARTWORK")
	f.hpbar.text:SetFontObject(GameFontHighlightSmall)
	f.hpbar.text:SetJustifyH("CENTER")

	f.mpbar = CreateFrame("StatusBar", nil, f)
	f.mpbar:SetStatusBarTexture(surface:Fetch(self.opt.Texture))
	f.mpbar:SetMinMaxValues(0,100)

	f.mpbar.texture = f.mpbar:CreateTexture(nil, "BORDER")
	f.mpbar.texture:SetTexture(surface:Fetch(self.opt.Texture))
	f.mpbar.texture:SetVertexColor(1, 0, 0, 0)

	f.mpbar.text = f.mpbar:CreateFontString(nil, "ARTWORK")
	f.mpbar.text:SetFontObject(GameFontHighlightSmall)
	f.mpbar.text:SetJustifyH("CENTER")

	self:SetStyle(f)

	f:Hide();

	tinsert(self.frames, f)
	
	ClickCastFrames = ClickCastFrames or {}
	ClickCastFrames[f] = true
end

function sRaidFrames:SetGroupFilters()
	for _, f in pairs(self.groupframes) do
		local id = f:GetID()
		if self.opt.SortBy == "group" then
			f:SetAttribute("groupFilter", id)
			f.title:SetText(string.format(L["Group %d"], id))
		elseif self.opt.SortBy == "class" then
			f:SetAttribute("groupFilter", self.classGroups[id])
			f.title:SetText(self.eclass2class[self.classGroups[id]])
		end
		f:Show()
	end
end

function sRaidFrames:SecureHeaderUpdate(frame)
	if frame.anchor then
		self:UpdateTitleVisibility(frame)
	end
end

function sRaidFrames:UpdateTitleVisibility(frame)
	local i = 1
	local hasChildren = false
	local child = frame:GetAttribute("child"..i)
	while child do
		if UnitExists(child.unit) then
			hasChildren = true
 			break
		end
		i = i + 1
		child = frame:GetAttribute("child"..i)
	end

	if sRaidFrames.opt.ShowGroupTitles and hasChildren then
		frame.anchor:Show()
	else
		frame.anchor:Hide()
	end
end

function sRaidFrames:SetSortMethods()
	for _, f in pairs(self.groupframes) do
		f:SetAttribute("sortMethod", "NAME")
		f:SetAttribute("sortDir", nil)
	end
end

function sRaidFrames:SetGrowth()
	for _, f in pairs(self.groupframes) do
		if self.opt.Growth == "up" then
			f:SetAttribute("point", "TOP")
			f.anchor:ClearAllPoints();
			f.anchor:SetPoint("TOP", f, "BOTTOM")
		elseif self.opt.Growth == "right" then
			f:SetAttribute("point", "RIGHT")
			f.anchor:ClearAllPoints();
			f.anchor:SetPoint("RIGHT", f, "LEFT")
		elseif self.opt.Growth == "left" then
			f:SetAttribute("point", "LEFT")
			f.anchor:ClearAllPoints();
			f.anchor:SetPoint("LEFT", f, "RIGHT")
		elseif self.opt.Growth == "down" then
			f:SetAttribute("point", "BOTTOM")
			f.anchor:ClearAllPoints();
			f.anchor:SetPoint("BOTTOM", f, "TOP")
		end
	end
	self:SetSpacing()
end

function sRaidFrames:SetSpacing()
	local s = sRaidFrames.opt.Spacing
	for _, f in pairs(self.groupframes) do
		if self.opt.Growth == "up" or self.opt.Growth == "down" then
			f:SetAttribute("xOffset", 0)
			f:SetAttribute("yOffset", s)
		elseif self.opt.Growth == "left" or self.opt.Growth == "right" then
			f:SetAttribute("xOffset", s)
			f:SetAttribute("yOffset", 0)
		end
	end
end

function sRaidFrames:StartMovingAll(baseFrame)
	this.multidrag = 1
	local id = baseFrame:GetID()
	local fg = self.groupframes[id]
	local x, y = fg:GetLeft(), fg:GetTop()
	if ( not x or not y ) then
		return
	end
	for k, f in pairs(self.groupframes) do
		if k ~= id then
			local oX, oY = f:GetLeft(), f:GetTop()
			if ( oX and oY ) then
				f:ClearAllPoints()
				f:SetPoint("TOPLEFT", fg, "TOPLEFT", oX-x, oY-y)
			end
		end
	end
end

function sRaidFrames:StopMovingOrSizingAll()
	this.multidrag = nil
	local id = this:GetID()
	local fg = self.groupframes[id]
	for k, f in pairs(self.groupframes) do
		if k ~= id then
			local oX, oY = f:GetLeft(), f:GetTop()
			if ( oX and oY ) then
				f:ClearAllPoints()
				f:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", oX, oY)
			end
		end
	end
end

function sRaidFrames:CreateGroupFrame(id)
	local f = CreateFrame("Frame", "sRaidGroupHeader" .. id, self.master, "SecureRaidGroupHeaderTemplate")
	f:SetAttribute("template", "sRaidFrameUnitSecure")
	f.initialConfigFunction = sRaidFrames_initialHeaderConfigFunction
	f:SetID(id)
	f:SetMovable(true)
		
	f.anchor = CreateFrame("Button", "sRaidFramesAnchor"..id, UIParent, "SecureUnitButtonTemplate")
	f.anchor:SetHeight(15)
	f.anchor:SetWidth(90)
	f.anchor:SetScript("OnDragStart", function() if self.opt.Locked then return end if IsAltKeyDown() then self:StartMovingAll(f) end f:StartMoving() end)
	f.anchor:SetScript("OnDragStop", function() if this.multidrag == 1 then self:StopMovingOrSizingAll() end f:StopMovingOrSizing(f) self:SavePosition() end)
	f.anchor:EnableMouse(true)
	f.anchor:RegisterForDrag("LeftButton")
	f.anchor:RegisterForClicks("LeftButtonUp","RightButtonUp","MiddleButtonUp","Button4Up","Button5Up")	
	
	f.title = f:CreateFontString(nil, "OVERLAY")
	f.title:SetParent(f.anchor)
	f.title:SetFontObject(GameFontNormal)
	f.title:SetJustifyH("CENTER")
	f.title:ClearAllPoints()
	f.title:SetPoint("CENTER", f.anchor)

	self.groupframes[id] = f
end

function sRaidFrames:SetBackdrop(f)
	if self.opt.Border then
		f:SetBackdrop({ bgFile = "Interface\\Tooltips\\UI-Tooltip-Background",
										tile = true,
										tileSize = 16,
										edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
										edgeSize = 16,
										insets = { left = 5, right = 5, top = 5, bottom = 5 }
									})
	else
		f:SetBackdrop({ bgFile = "Interface\\Tooltips\\UI-Tooltip-Background",
										tile = true, tileSize = 16,
										insets = { left = 0, right = 0, top = 0, bottom = 0 }
									})
	end

	f:SetBackdropColor(self.opt.BackgroundColor.r, self.opt.BackgroundColor.g, self.opt.BackgroundColor.b, self.opt.BackgroundColor.a)
	f:SetBackdropBorderColor(self.opt.BorderColor.r, self.opt.BorderColor.g, self.opt.BorderColor.b, self.opt.BorderColor.a)
end

function sRaidFrames:SetStyle(f)
	self:SetWHP(f, 90, 40)
	self:SetWHP(f.title, 80, 16, "TOPLEFT", f, "TOPLEFT",  5, -4)
	self:SetWHP(f.aura1, 16, 16, "TOPRIGHT", f, "TOPRIGHT", -4, -4)
	self:SetWHP(f.aura2, 16, 16, "RIGHT", f.aura1, "LEFT", 0, 0)
	self:SetWHP(f.buff1, 12, 12, "TOPRIGHT", f, "TOPRIGHT", -4, -4)
	self:SetWHP(f.buff2, 12, 12, "RIGHT", f.buff1, "LEFT", 0, 0)
	self:SetWHP(f.buff3, 12, 12, "RIGHT", f.buff2, "LEFT", 0, 0)
	self:SetWHP(f.buff4, 12, 12, "RIGHT", f.buff3, "LEFT", 0, 0)
	self:SetWHP(f.hpbar, 80, 12, "TOPLEFT", f.title, "BOTTOMLEFT", 0, 0)
	self:SetWHP(f.mpbar, 80, 4, "TOPLEFT", f.hpbar, "BOTTOMLEFT", 0, 0)

	self:SetWHP(f.hpbar.text, f.hpbar:GetWidth(), f.hpbar:GetHeight(), "CENTER", f.hpbar, "CENTER", 0, 0)
	self:SetWHP(f.mpbar.text, f.mpbar:GetWidth(), f.mpbar:GetHeight(), "CENTER", f.mpbar, "CENTER", 0, 0)

	self:SetBackdrop(f)
end

function sRaidFrames:SetWHP(frame, width, height, p1, relative, p2, x, y)
	frame:SetWidth(width)
	frame:SetHeight(height)

	if (p1) then
		frame:ClearAllPoints()
		frame:SetPoint(p1, relative, p2, x, y)
	end
end

function sRaidFrames:FindUnitFrames(unit)
	local res = {}
	for _, f in ipairs(self.frames) do
		if f.unit and f.unit == unit then
			tinsert(res, f)
		end
	end
	return res
end

function sRaidFrames:SetPosition()
	if not self.opt.Positions then
		self:ResetPosition()
	else
		self:RestorePosition()
	end
end

function sRaidFrames:SavePosition()
	local aryPos = {}
	local s = self.master:GetEffectiveScale()

	for k,f in pairs(self.groupframes) do
		aryPos[k] = {x = f:GetLeft()*s, y = f:GetTop()*s}
	end

	self:S("Positions", aryPos)
end

function sRaidFrames:RestorePosition()
	local aryPos = self.opt.Positions
	local s = self.master:GetEffectiveScale()

	for k, pos in pairs(aryPos) do
		local f = self.groupframes[k]
		f:ClearAllPoints()
		f:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", pos["x"]/s, pos["y"]/s)
	end
end

function sRaidFrames:ResetPosition()
	self:PositionLayout("ctra", 200, -200)
end

function sRaidFrames:PositionLayout(layout, xBuffer, yBuffer)
	local xMod, yMod, i = 0, 0, -1
	local frameHeight = 40+3 -- TODO

	for k,v in pairs(self.groupframes) do
		i = i + 1
		if layout == "horizontal" then
			yMod = (i) * v:GetWidth()
			xMod = 0
		elseif layout == "vertical" then
			if i ~= 0 and math_mod(i, 2) == 0 then
				xMod = xMod + (-1*MEMBERS_PER_RAID_GROUP*frameHeight)
				yMod = 0
				i = 0
			else
				yMod = i * v:GetWidth()
			end
		elseif layout == "ctra" then
			if i ~= 0 and math_mod(i, 2) == 0 then
				yMod = yMod + v:GetWidth()
				xMod = 0
				i = 0
			else
				xMod = i * (-1*MEMBERS_PER_RAID_GROUP*frameHeight)
			end
		end

		v:ClearAllPoints()
		v:SetPoint("TOPLEFT", UIParent, "TOPLEFT", xBuffer+yMod, yBuffer+xMod)
	end

	self:SavePosition()
end

local Tablet = AceLibrary("Tablet-2.0")
function sRaidFrames:OnTooltipUpdate()
	Tablet:SetHint(L["Right-click for options."])
end

function sRaidFrames:S(var, val)
	self.db.profile[var] = val
end
